use std::io::Read;
use std::fs::read_to_string;
use std::sync::{Mutex, RwLock};
use std::collections::VecDeque;
use rustorm::{pool::Pool, DbError};
use chrono::offset::Utc;

mod config;
use config::{Config, CalendarOptions};
mod schema;
use schema::{new, Calendar, Event};
mod transaction;
use transaction::{Transaction, QuerySql};
mod ics;
use ics::Parser;

pub struct Resources {
    db_pool: Mutex<Pool>,
    db_url: String,
    http_client: reqwest::Client,
    queue: RwLock<VecDeque<(String, CalendarOptions)>>,
}

impl Resources {
    pub fn new(db_url: String, queue: VecDeque<(String, CalendarOptions)>) -> Self {
        let db_pool = Mutex::new(Pool::new());
        let http_client = reqwest::Client::new();
        let queue = RwLock::new(queue);
        Resources { db_pool, db_url, http_client, queue }
    }

    fn transaction(&self) -> Result<Transaction, DbError> {
        let em = self.db_pool.lock()
            .unwrap()
            .em(&self.db_url)?;
        Transaction::new(em)
    }

    fn next(&self) -> Option<(String, CalendarOptions)> {
        let mut queue = self.queue.write().unwrap();
        queue.pop_front()
    }

    fn worker(&self) -> Result<(), Error> {
        let (id, cal_opts) = match self.next() {
            Some(next) => next,
            None => return Ok(()),
        };
        let tx = self.transaction()?;

        let cal: Option<Calendar> =
            tx.query("SELECT * FROM calendar WHERE id=$1", &[&id])?;
        if cal.is_none() {
            tx.insert(&[&new::Calendar {
                id: id.clone(),
                url: cal_opts.url.clone(),
                last_fetch: Utc::now(),
            }])?;
        } else {
            tx.exec("UPDATE calendar SET last_fetch=$2 WHERE id=$1", &[&id, &"NOW()"])?;
        }

        let mut res = self.http_client.get(&cal_opts.url).send()?;
        println!("{} {} {:?}", res.status(), cal_opts.url, res.headers());

        let mut p = Parser::new();
        let mut buf = [0; 1024];
        loop {
            match res.read(&mut buf)? {
                len if len > 0 => {
                    let data = &buf[..len];
                    p.feed(data, |obj| {
                        println!("- [{}] {}", obj.get("DTSTART").unwrap_or("?"), obj.get("SUMMARY").unwrap_or("?"));
                        print!("  {}", obj.get("LOCATION").unwrap_or("?"));
                        obj.get("URL").map(|url| print!("  <{}>", url));
                        println!("");
                    });
                }
                _ => break,
            }
        }

        tx.commit()?;
        
        Ok(())
    }
}

#[derive(Debug)]
pub enum Error {
    Db(DbError),
    Http(reqwest::Error),
    Io(std::io::Error),
}

impl From<DbError> for Error {
    fn from(e: DbError) -> Self {
        Error::Db(e)
    }
}

impl From<reqwest::Error> for Error {
    fn from(e: reqwest::Error) -> Self {
        Error::Http(e)
    }
}

impl From<std::io::Error> for Error {
    fn from(e: std::io::Error) -> Self {
        Error::Io(e)
    }
}

fn main() {
    let config_file = read_to_string("config.yaml")
        .expect("config.yaml");
    let config: Config = serde_yaml::from_str(&config_file)
        .expect("config");
    let res = Resources::new(
        config.db_url,
        config.calendars.into_iter()
            .collect()
    );

    crossbeam::scope(|s| {
        let cpus = num_cpus::get();
        let workers: Vec<_> = (0..cpus)
            .map(|_| s.spawn(|_| {
                res.worker()
                    .map_err(|e| println!("{:?}", e))
            }))
            .collect();
        for worker in workers.into_iter() {
            worker.join().expect("worker").unwrap();
        }
    }).unwrap();
}
